@@ -20,6 +20,7 @@ from .astutil import Comment
 from .astutil import Symbol
 from .astutil import Builtin
 from .astutil import Static
+from .astutil import TokenRef
 
 from .codegen import TemplateCodeGenerator
 from .codegen import template
@@ -54,7 +55,6 @@ from .utils import safe_native
 from .utils import builtins
 from .utils import decode_htmlentities
 
-
 if version >= (3, 0, 0):
     long = int
 
@@ -108,7 +108,13 @@ def store_rcontext(name):
     return subscript(name, load("rcontext"), ast.Store())
 
 
-def set_error(token, exception):
+def set_token(stmts, token):
+    pos = getattr(token, "pos", 0)
+    body = template("__token = pos", pos=TokenRef(pos, len(token)))
+    return body + stmts
+
+
+def eval_token(token):
     try:
         line, column = token.location
         filename = token.filename
@@ -119,57 +125,42 @@ def set_error(token, exception):
     string = safe_native(token)
 
     return template(
-        "rcontext.setdefault('__error__', [])."
-        "append((string, line, col, src, exc))",
+        "(string, line, col)",
         string=ast.Str(s=string),
         line=ast.Num(n=line),
         col=ast.Num(n=column),
-        src=ast.Str(s=filename),
-        sys=Symbol(sys),
-        exc=exception,
-        )
+        mode="eval"
+    )
 
 
-def try_except_wrap(stmts, token):
-    exception = template(
-        "exc_info()[1]", exc_info=Symbol(sys.exc_info), mode="eval"
-        )
-
-    body = set_error(token, exception) + template("raise")
-
-    return ast.TryExcept(
-        body=stmts,
-        handlers=[ast.ExceptHandler(body=body)],
-        )
+emit_node = template(is_func=True, func_args=('node',), source=r"""
+    __append(node)""")
 
 
-@template
-def emit_node(node):  # pragma: no cover
-    __append(node)
-
-
-@template
-def emit_node_if_non_trivial(node):  # pragma: no cover
+emit_node_if_non_trivial = template(is_func=True, func_args=('node',),
+                                    source=r"""
     if node is not None:
         __append(node)
+""")
 
 
-@template
-def emit_bool(target, s, default_marker=None,
-                 default=None):  # pragma: no cover
+emit_bool = template(is_func=True,
+                     func_args=('target', 's', 'default_marker', 'default'),
+                     func_defaults=(None, None), source=r"""
     if target is default_marker:
         target = default
     elif target:
         target = s
     else:
-        target = None
+        target = None""")
 
 
-@template
-def emit_convert(
-    target, encoded=byte_string, str=unicode_string,
-    long=long, type=type,
-    default_marker=None, default=None):  # pragma: no cover
+emit_convert = template(is_func=True,
+                        func_args=('target', 'encoded', 'str', 'long', 'type',
+                                   'default_marker', 'default'),
+                        func_defaults=(byte_string, unicode_string, long, type,
+                                       None),
+                        source=r"""
     if target is None:
         pass
     elif target is default_marker:
@@ -188,13 +179,14 @@ def emit_convert(
                 __converted = convert(target)
                 target = str(target) if target is __converted else __converted
             else:
-                target = target()
+                target = target()""")
 
 
-@template
-def emit_func_convert(
-    func, encoded=byte_string, str=unicode_string,
-    long=long, type=type):  # pragma: no cover
+emit_func_convert = template(is_func=True,
+                             func_args=('func', 'encoded', 'str','long','type'),
+                             func_defaults=(byte_string, unicode_string, long,
+                                            type),
+                             source=r"""
     def func(target):
         if target is None:
             return
@@ -216,20 +208,24 @@ def emit_func_convert(
             else:
                 target = target()
 
-        return target
+        return target""")
 
 
-@template
-def emit_translate(target, msgid, target_language, default=None):  \
-    # pragma: no cover
-    target = translate(msgid, default=default, domain=__i18n_domain, context=__i18n_context, target_language=target_language)
+emit_translate = template(is_func=True,
+                          func_args=('target', 'msgid', 'target_language',
+                          'default'),
+                          func_defaults=(None,),
+                          source=r"""
+    target = translate(msgid, default=default, domain=__i18n_domain, 
+                       context=__i18n_context, 
+                       target_language=target_language)""")
 
 
-@template
-def emit_func_convert_and_escape(
-    func, str=unicode_string, long=long,
-    type=type, encoded=byte_string):  # pragma: no cover
-
+emit_func_convert_and_escape = template(
+    is_func=True,
+    func_args=('func', 'str', 'long', 'type', 'encoded'),
+    func_defaults=(unicode_string, long, type, byte_string,),
+    source=r"""
     def func(target, quote, quote_entity, default, default_marker):
         if target is None:
             return
@@ -271,7 +267,7 @@ def emit_func_convert_and_escape(
                         if quote is not None and quote in target:
                             target = target.replace(quote, quote_entity)
 
-        return target
+        return target""")
 
 
 class Interpolator(object):
@@ -283,11 +279,13 @@ class Interpolator(object):
         r'(?<![\\$])\$({(?P<expression>.*)}|(?P<variable>[A-Za-z][A-Za-z0-9_]*))',
         re.DOTALL)
 
-    def __init__(self, expression, braces_required, translate=False):
+    def __init__(self, expression, braces_required, translate=False,
+                 decode_htmlentities=False):
         self.expression = expression
         self.regex = self.braces_required_regex if braces_required else \
                      self.braces_optional_regex
         self.translate = translate
+        self.decode_htmlentities = decode_htmlentities
 
     def __call__(self, name, engine):
         """The strategy is to find possible expression strings and
@@ -343,7 +341,9 @@ class Interpolator(object):
             while True:
                 d = groupdict(m, matched)
                 string = d["expression"] or d.get("variable") or ""
-                string = decode_htmlentities(string)
+
+                if self.decode_htmlentities:
+                    string = decode_htmlentities(string)
 
                 if string:
                     try:
@@ -385,7 +385,7 @@ class Interpolator(object):
                 target = template(
                     "translate(msgid, domain=__i18n_domain, context=__i18n_context, target_language=target_language)",
                     msgid=target, mode="eval",
-                    target_language="target_language",
+                    target_language=load("target_language"),
                     )
         else:
             if translate:
@@ -405,18 +405,19 @@ class Interpolator(object):
                 target = template(
                     "translate(msgid, mapping=mapping, domain=__i18n_domain, context=__i18n_context, target_language=target_language)",
                     msgid=ast.Str(s=formatting_string),
-                    target_language=target_language,
+                    target_language=load("target_language"),
                     mapping=ast.Dict(keys=keys, values=values),
                     mode="eval"
                     )
             else:
                 nodes = [
+                    node if isinstance(node, ast.Str) else
                     template(
                         "NODE if NODE is not None else ''",
                         NODE=node, mode="eval"
-                        )
+                    )
                     for node in nodes
-                    ]
+                ]
 
                 target = ast.BinOp(
                     left=ast.Str(s="%s" * len(nodes)),
@@ -492,28 +493,30 @@ class ExpressionEngine(object):
         compiler = self.parse(string)
         return compiler(string, target)
 
-    def parse(self, string, handle_errors=True):
+    def parse(self, string, handle_errors=True, char_escape=None):
         expression = self._parser(string)
-        compiler = self.get_compiler(expression, string, handle_errors)
+        compiler = self.get_compiler(expression, string, handle_errors, char_escape)
         return ExpressionCompiler(compiler, self)
 
-    def get_compiler(self, expression, string, handle_errors):
+    def get_compiler(self, expression, string, handle_errors, char_escape):
+        if char_escape is None:
+            char_escape = self._char_escape
         def compiler(target, engine, result_type=None, *args):
             stmts = expression(target, engine)
 
             if result_type is not None:
                 method = getattr(self, '_convert_%s' % result_type)
-                steps = method(target, *args)
+                steps = method(target, char_escape, *args)
                 stmts.extend(steps)
 
             if handle_errors:
-                return [try_except_wrap(stmts, string.strip())]
+                return set_token(stmts, string.strip())
 
             return stmts
 
         return compiler
 
-    def _convert_bool(self, target, s):
+    def _convert_bool(self, target, char_escape, s):
         """Converts value given by ``target`` to a string ``s`` if the
         target is a true value, otherwise ``None``.
         """
@@ -524,36 +527,8 @@ class ExpressionEngine(object):
             default_marker=self._default_marker
             )
 
-    def _convert_text(self, target):
-        """Converts value given by ``target`` to text."""
-
-        if self._char_escape:
-            # This is a cop-out - we really only support a very select
-            # set of escape characters
-            other = set(self._char_escape) - self.supported_char_escape_set
-
-            if other:
-                for supported in '"', '\'', '':
-                    if supported in self._char_escape:
-                        quote = supported
-                        break
-                else:
-                    raise RuntimeError(
-                        "Unsupported escape set: %s." % repr(self._char_escape)
-                        )
-            else:
-                quote = '\0'
-
-            entity = char2entity(quote or '\0')
-
-            return template(
-                "TARGET = __quote(TARGET, QUOTE, Q_ENTITY, DEFAULT, MARKER)",
-                TARGET=target,
-                QUOTE=ast.Str(s=quote),
-                Q_ENTITY=ast.Str(s=entity),
-                DEFAULT=self._default,
-                MARKER=self._default_marker,
-                )
+    def _convert_structure(self, target, char_escape):
+        """Converts value given by ``target`` to structure output."""
 
         return emit_convert(
             target,
@@ -561,6 +536,39 @@ class ExpressionEngine(object):
             default_marker=self._default_marker,
             )
 
+    def _convert_text(self, target, char_escape):
+        """Converts value given by ``target`` to text."""
+
+        if not char_escape:
+            return self._convert_structure(target, char_escape)
+
+        # This is a cop-out - we really only support a very select
+        # set of escape characters
+        other = set(char_escape) - self.supported_char_escape_set
+
+        if other:
+            for supported in '"', '\'', '':
+                if supported in char_escape:
+                    quote = supported
+                    break
+            else:
+                raise RuntimeError(
+                    "Unsupported escape set: %s." % repr(char_escape)
+                    )
+        else:
+            quote = '\0'
+
+        entity = char2entity(quote or '\0')
+
+        return template(
+            "TARGET = __quote(TARGET, QUOTE, Q_ENTITY, DEFAULT, MARKER)",
+            TARGET=target,
+            QUOTE=ast.Str(s=quote),
+            Q_ENTITY=ast.Str(s=entity),
+            DEFAULT=self._default,
+            MARKER=self._default_marker,
+            )
+
 
 class ExpressionCompiler(object):
     def __init__(self, compiler, engine):
@@ -633,8 +641,9 @@ class ExpressionEvaluator(object):
             module = Module("evaluate", Context(assignment))
 
             compiler = Compiler(
-                self._engine, module, ('econtext', 'rcontext') + self._names
-                )
+                self._engine, module, "<string>", string,
+                ('econtext', 'rcontext') + self._names
+            )
 
             env = {}
             exec(compiler.code, env)
@@ -755,12 +764,9 @@ class ExpressionTransform(object):
 
             stmts = template(
                 "__exc = loads(p)", loads=self.loads_symbol, p=ast.Str(s=p)
-                )
+            )
 
-            token = Token(exc.token, exc.offset, filename=exc.filename)
-
-            stmts += set_error(token, load("__exc"))
-            stmts += [ast.Raise(exc=load("__exc"))]
+            stmts += set_token([ast.Raise(exc=load("__exc"))], exc.token)
 
         # Apply visitor to each statement
         for stmt in stmts:
@@ -808,11 +814,8 @@ class ExpressionTransform(object):
         return [ast.Assign(targets=[target], value=value)]
 
     def visit_Substitution(self, node, target):
-        engine = self.engine_factory(
-            char_escape=node.char_escape,
-            default=node.default,
-            )
-        compiler = engine.parse(node.value)
+        engine = self.engine_factory(default=node.default)
+        compiler = engine.parse(node.value, char_escape=node.char_escape)
         return compiler.assign_text(target)
 
     def visit_Negate(self, node, target):
@@ -851,11 +854,15 @@ class ExpressionTransform(object):
             raise RuntimeError("Bad value: %r." % node.value)
 
         interpolator = Interpolator(
-            expr.value, node.braces_required, node.translation
-            )
+            expr.value, node.braces_required,
+            translate=node.translation,
+            decode_htmlentities=True
+        )
 
-        compiler = engine.get_compiler(interpolator, expr.value, True)
-        return compiler(target, engine)
+        compiler = engine.get_compiler(
+            interpolator, expr.value, True, ()
+        )
+        return compiler(target, engine, "text")
 
     def visit_Translate(self, node, target):
         if node.msgid is not None:
@@ -900,7 +907,8 @@ class Compiler(object):
 
     global_builtins = set(builtins.__dict__)
 
-    def __init__(self, engine_factory, node, builtins={}, strict=True):
+    def __init__(self, engine_factory, node, filename, source,
+                 builtins={}, strict=True):
         self._scopes = [set()]
         self._expression_cache = {}
         self._translations = []
@@ -937,14 +945,27 @@ class Compiler(object):
             module = ast.Module([])
             module.body += self.visit(node)
             ast.fix_missing_locations(module)
-            generator = TemplateCodeGenerator(module)
+            prelude = "__filename = %r" % filename
+            generator = TemplateCodeGenerator(module, source)
+            tokens = [
+                Token(source[pos:pos + length], pos, source)
+                for pos, length in generator.tokens
+            ]
+            token_map_def = "__tokens = {" + ", ".join("%d: %r" % (
+                token.pos,
+                (token, ) + token.location
+            ) for token in tokens) + "}"
         finally:
             if backup is not None:
                 node_annotations.clear()
                 node_annotations.update(backup)
                 self.lock.release()
 
-        self.code = generator.code
+        self.code = "\n".join((
+            prelude,
+            token_map_def,
+            generator.code
+        ))
 
     def visit(self, node):
         if node is None:
@@ -987,7 +1008,7 @@ class Compiler(object):
 
         body += template(
             r"__re_whitespace = "
-            r"functools.partial(re.compile('\s+').sub, ' ')",
+            r"functools.partial(re.compile('\\s+').sub, ' ')",
         )
 
         # Visit module content
@@ -1034,6 +1055,7 @@ class Compiler(object):
         # Initialization
         body += template("__append = __stream.append")
         body += template("__re_amp = g_re_amp")
+        body += template("__token = None")
         body += template("__re_needs_escape = g_re_needs_escape")
 
         body += emit_func_convert("__convert")
@@ -1059,8 +1081,25 @@ class Compiler(object):
                 "except: NAME = None",
                 KEY=ast.Str(s=name), NAME=store(name))
 
-        # Append visited nodes
-        body += nodes
+        exc = template(
+            "exc_info()[1]", exc_info=Symbol(sys.exc_info), mode="eval"
+        )
+
+        exc_handler = template(
+            "if pos is not None: rcontext.setdefault('__error__', [])."
+            "append(token + (__filename, exc, ))",
+            exc=exc,
+            token=template("__tokens[pos]", pos="__token", mode="eval"),
+            pos="__token"
+        ) + template("raise")
+
+        # Wrap visited nodes in try-except error handler.
+        body += [
+            ast.TryExcept(
+                body=nodes,
+                handlers=[ast.ExceptHandler(body=exc_handler)]
+            )
+        ]
 
         function_name = "render" if node.name is None else \
                         "render_%s" % mangle(node.name)
@@ -1109,7 +1148,7 @@ class Compiler(object):
         self._leave_assignment((node.name, ))
 
         error_assignment = template(
-            "econtext[key] = cls(__exc, rcontext['__error__'][-1][1:3])",
+            "econtext[key] = cls(__exc, __tokens[__token][1:3])",
             cls=ErrorInfo,
             key=ast.Str(s=node.name),
             )
@@ -1454,8 +1493,8 @@ class Compiler(object):
             render = "render"
         else:
             render = "render_%s" % mangle(node.name)
-
-        return template(
+        token_reset = template("__token = None")
+        return token_reset + template(
             "f(__stream, econtext.copy(), rcontext, __i18n_domain)",
             f=render) + \
             template("econtext.update(rcontext)")
@@ -1518,7 +1557,7 @@ class Compiler(object):
         for stmt in stmts:
             self._visitor(stmt)
 
-        return [try_except_wrap(stmts, node.source)]
+        return set_token(stmts, node.source)
 
     def visit_UseExternalMacro(self, node):
         self._macros.append(node.extend)
@@ -1577,10 +1616,10 @@ class Compiler(object):
         return (
             callbacks +
             assignment +
-            [try_except_wrap(
+            set_token(
                 template("__m = __macro.include"),
                 node.expression.value
-            )] +
+            ) +
             template(
                 "__m(__stream, econtext.copy(), "
                 "rcontext, __i18n_domain)"
